import typing
from dataclasses import dataclass, field


class ChartingException(Exception):
    pass


@dataclass
class Color:
    color: str

    def __post_init__(self):
        """
        Color must be strictly a 6-character hex string
        """
        try:
            if not (
                len(self.color) == 6 and int(self.color, 16) > 0
            ):  # it is fine without regular expr
                raise AssertionError
        except AssertionError:
            raise ChartingException(
                "Color must be strictly a 6-character hex string. "
                f"The value '{self.color}' is not a such."
            )

    def as_hex(self) -> str:
        return f"#{self.color}"


@dataclass
class __InheritableDataclassAllowingSuperPostInit:
    """
    Just a simple dataclass defining a useless __post_init__ in order to satisfy python's MRO and allow inherited
    dataclasses with custom __post_init__'s
    """

    def __post_init__(self):
        pass


@dataclass
class Colored(__InheritableDataclassAllowingSuperPostInit):
    """
    Used to assign the color to the component of the diagram.
    """

    _color: typing.Optional[str]
    color: typing.Optional[Color] = field(init=False)

    def __post_init__(self):
        super().__post_init__()
        self.color = Color(self._color) if self._color else None


class BaseChart:
    """
    Base class for all the charts
    """



import typing

from umlcharter.charts.sequence_diagram import (
    SequenceDiagram,
    SequenceDiagramParticipant,
    SequenceDiagramParticipantGroup,
    Step,
    ParticipantActivationControl,
    ForwardStep,
    GroupControl,
    LoopControl,
    ConditionControl,
    CaseControl,
    ReturnStep,
    NoteStep,
)


class D2SequenceDiagram:
    @staticmethod
    def _line_break(string: str) -> str:
        """Some places allow line break as \n"""
        return string.replace("\n", "\\n") or "''"

    @classmethod
    def generate(cls, sequence_diagram: SequenceDiagram) -> str:
        participants: typing.Dict[
            SequenceDiagramParticipantGroup, typing.List[SequenceDiagramParticipant]
        ] = sequence_diagram._SequenceDiagram__participants  # noqa
        sequence: typing.List[Step] = sequence_diagram._SequenceDiagram__sequence  # noqa

        last_targeted_participant: SequenceDiagramParticipant | None = None
        aliases: typing.Dict[SequenceDiagramParticipant, str] = {}
        aliases_counter: int = 1

        participant_types_map = {
            "default": "",
            "actor": "person",
            "boundary": "",
            "control": "",
            "entity": "",
        }

        generated = f"title: {cls._line_break(sequence_diagram.title)} {{\nshape: sequence_diagram\n"
        for _, group_participants in participants.items():
            for participant in group_participants:
                # define initial last targeted participant
                if not last_targeted_participant:
                    last_targeted_participant = participant

                aliases[participant] = f"p{aliases_counter}"
                aliases_counter += 1

                generated += (
                    f"{aliases[participant]}: {cls._line_break(participant.title)} "
                )
                if participant.color or participant_types_map[participant.type_]:
                    generated += "{\n"
                    if participant.color:
                        generated += (
                            f'style: {{fill: "{participant.color.as_hex()}" \n'
                            f'stroke:"{participant.color.as_hex()}" }}\n'
                        )
                    if participant_types_map[participant.type_]:
                        generated += (
                            f"shape: {participant_types_map[participant.type_]}\n"
                        )
                    generated += "}"
                generated += "\n"

        # NB! In D2 the logic of "activation" phases or "spans" works a bit differently, compared to the other DSLs.
        # You have to know that the participant will be activated
        # BEFORE the flow goes under the new participant control.
        # To align the activation representation with the other DSLs, we have to
        # 1. Check if the step is `ForwardStep` to the participant X
        # 2. If right after the `ForwardStep` participant X is activated -
        #   swap these `ForwardStep` & `ParticipantActivationControl`
        new_seq = sequence.copy()
        len_sequence = len(new_seq)
        for index, step in enumerate(new_seq):
            if (
                isinstance(step, ForwardStep)
                and (index + 1) < len_sequence
                and isinstance(new_seq[index + 1], ParticipantActivationControl)
                and new_seq[index + 1].is_active is True
                and step.to_participant is new_seq[index + 1].participant
            ):
                new_seq[index], new_seq[index + 1] = new_seq[index + 1], new_seq[index]

        activation_counter = 0
        custom_element_counter = 1
        for step in new_seq:
            if isinstance(step, ParticipantActivationControl):
                if step.is_active:
                    aliases[step.participant] += f".{activation_counter}"
                    activation_counter += 1
                else:
                    aliases[step.participant] = ".".join(
                        aliases[step.participant].split(".")[:-1]
                    )

            if isinstance(step, ForwardStep):
                generated += (
                    f"{aliases[step.from_participant]} -> "
                    f"{aliases[step.to_participant]}: {cls._line_break(step.text)}\n"
                )
                last_targeted_participant = step.to_participant

            if isinstance(step, ReturnStep):
                generated += (
                    f"{aliases[step.from_participant]} -> "
                    f"{aliases[step.to_participant]}: {cls._line_break(step.text)} {{style.stroke-dash: 3}}\n"
                )
                last_targeted_participant = step.to_participant

            if isinstance(step, NoteStep):
                generated += f'{aliases[last_targeted_participant]}."{cls._line_break(step.text)}"\n'

            if isinstance(step, GroupControl):
                if step.is_active:
                    generated += rf"group{custom_element_counter}: \[GROUP\] {cls._line_break(step.text)}: {{"
                    if step.color:
                        generated += f'\nstyle: {{\nfill: "{step.color.as_hex()}" \n}}'
                    generated += "\n"
                    custom_element_counter += 1
                else:
                    generated += "}\n"

            if isinstance(step, LoopControl):
                if step.is_active:
                    generated += rf"loop{custom_element_counter}: \[LOOP\] {cls._line_break(step.how_many_iterations)}: {{"
                    if step.color:
                        generated += f'\nstyle: {{\nfill: "{step.color.as_hex()}" \n}}'
                    generated += "\n"
                    custom_element_counter += 1
                else:
                    generated += "}\n"

            if isinstance(step, ConditionControl):
                if step.is_active:
                    generated += rf"alt{custom_element_counter}: \[ALT\] {{"
                    if step.color:
                        generated += f'\nstyle: {{\nfill: "{step.color.as_hex()}" \n}}'
                    generated += "\n"
                    custom_element_counter += 1
                else:
                    generated += "}\n"

            if isinstance(step, CaseControl):
                if step.is_active:
                    generated += rf"case{custom_element_counter}: \[CASE\] {cls._line_break(step.text)}: {{"
                    if step.color:
                        generated += f'\nstyle: {{\nfill: "{step.color.as_hex()}" \n}}'
                    generated += "\n"
                    custom_element_counter += 1
                else:
                    generated += "}\n"

        return generated + "}\n"



from .charts.sequence_diagram import SequenceDiagram
from .charts.graph_diagram import GraphDiagram
from .generators.mermaid.mermaid import Mermaid
from .generators.plantuml.plantuml import PlantUML
from .generators.d2.d2 import D2
from .generators.sequencediagramorg.sequencediagramorg import SequenceDiagramOrg

__version__ = "1.1.2"

__all__ = (
    # charts
    "SequenceDiagram",
    "GraphDiagram",
    # generators
    "Mermaid",
    "PlantUML",
    "D2",
    "SequenceDiagramOrg",
)



import typing
from dataclasses import dataclass, field

from umlcharter.charts.common import BaseChart, Colored, ChartingException
from umlcharter.generators.base import IChartGenerator


@dataclass
class BaseNode:
    _graph_ref: typing.Optional["Node"]

    @property
    def __graph_belongs_to(
        self,
    ) -> typing.Dict["BaseNode", typing.List[typing.Tuple["BaseNode", str]]]:
        return self._graph_ref._Node__inner_graph  # noqa

    def __check_if_interaction_is_allowed(self, to: "BaseNode"):
        if to not in self.__graph_belongs_to:
            raise ChartingException(
                "You cannot define a link from a node to another one outside of the same level / group"
            )

        for already_existing_routes in self.__graph_belongs_to[self]:
            if already_existing_routes[0] is to:
                raise ChartingException(
                    f"There is already an established link from {self} to {to}."
                )

        if isinstance(self, Start) and isinstance(to, Finish):
            raise ChartingException(
                "The direct link from start to finish is pointless, these are abstract nodes;"
                "there must be some intermediate nodes between them."
            )

        if isinstance(self, Finish):
            raise ChartingException(
                "The 'finish' node can only be the destination and not the starting point"
            )

        if isinstance(to, Start):
            raise ChartingException(
                "The 'start' node can only be the starting point and not the destination"
            )

    def go_to(self, to: "BaseNode", text: str = "") -> "BaseNode":
        self.__check_if_interaction_is_allowed(to)
        self.__graph_belongs_to[self].append((to, text))
        return to


@dataclass
class Fork(BaseNode):
    _notes: list[str] = field(default_factory=list)

    def __hash__(self):
        return hash(f"{id(self)}")

    def __repr__(self):
        return "Fork"  # pragma: nocover

    def note(self, text: str) -> None:
        self._notes.append(text)


@dataclass
class Join(BaseNode):
    _notes: list[str] = field(default_factory=list)

    def __hash__(self):
        return hash(f"{id(self)}")

    def __repr__(self):
        return "Join"  # pragma: nocover

    def note(self, text: str) -> None:
        self._notes.append(text)


@dataclass
class Condition(BaseNode):
    _notes: list[str] = field(default_factory=list)

    def __hash__(self):
        return hash(f"{id(self)}")

    def __repr__(self):
        return "Condition"  # pragma: nocover

    def note(self, text: str) -> None:
        self._notes.append(text)


@dataclass
class Start(BaseNode):
    def __hash__(self):
        return hash(f"{id(self)}")

    def __repr__(self):
        return "Start"  # pragma: nocover


@dataclass
class Finish(BaseNode):
    def __hash__(self):
        return hash(f"{id(self)}")

    def __repr__(self):
        return "Finish"  # pragma: nocover


@dataclass
class Node(BaseNode, Colored):
    text: str
    _notes: list[str] = field(default_factory=list)
    start: Start = field(init=False)
    finish: Finish = field(init=False)
    __inner_graph: typing.Dict[BaseNode, typing.List[typing.Tuple[BaseNode, str]]] = (
        field(init=False)
    )

    def is_group(self) -> bool:
        """The node can be a representation of a group / composite state if it contains the other nodes inside it."""
        return (
            len(self.__inner_graph) > 2
        )  # contains anything besides the default start and end nodes

    def is_top_level(self) -> bool:
        return not self._graph_ref

    def __hash__(self):
        return hash(f"{id(self)}")

    def __repr__(self):
        return (
            f"{'Group' if self.is_group() else 'Node'} {self.text}"  # pragma: nocover
        )

    def __post_init__(self):
        super().__post_init__()
        self.start = Start(_graph_ref=self)
        self.finish = Finish(_graph_ref=self)
        self.__inner_graph = {
            self.start: [],
            self.finish: [],
        }
        self._notes = []

    def __check_if_adding_new_element_is_allowed(self, title: str):
        for nodes in self.__inner_graph:
            if getattr(nodes, "text", None) == title:
                raise ChartingException(
                    f"There must be no nodes in the graph in the same group with the same title '{title}'."
                )

    def node(self, title: str, color: typing.Optional[str] = None) -> "Node":
        self.__check_if_adding_new_element_is_allowed(title)
        node = Node(_graph_ref=self, text=title, _color=color)
        self.__inner_graph[node] = []
        return node

    def fork(self) -> "Fork":
        fork = Fork(_graph_ref=self)
        self.__inner_graph[fork] = []
        return fork

    def join(self) -> "Join":
        join = Join(_graph_ref=self)
        self.__inner_graph[join] = []
        return join

    def condition(self) -> "Condition":
        condition = Condition(_graph_ref=self)
        self.__inner_graph[condition] = []
        return condition

    def note(self, text: str) -> None:
        self._notes.append(text)


@dataclass
class GraphDiagram(BaseChart):
    """
    A graph that must be rendered to the diagram DSL - depending on the chosen generator.

    :title: The title of the diagram to display on the top of the diagram
    :generator_cls: the class of the generator to be used to generate the diagram
    :is_vertical: The boolean flag used to define the default orientation of the rendered graph.
        If it is set to True, the orientation of the diagram is set to render the nodes from top to bottom.
        Otherwise, the orientation of the diagram is set to render the nodes from left to right.
        True by default (renders from top to bottom)
    """

    title: str
    generator_cls: typing.Type[IChartGenerator]
    is_vertical: bool = True

    __generator: IChartGenerator = field(init=False)
    __base_node: Node = field(init=False)

    def __post_init__(self):
        self.__generator = self.generator_cls(self)
        self.__base_node = Node(_graph_ref=None, text="", _color=None)

    @property
    def start(self) -> Start:
        return self.__base_node.start

    @property
    def finish(self) -> Finish:
        return self.__base_node.finish

    def node(self, title: str, color: typing.Optional[str] = None) -> Node:
        return self.__base_node.node(title, color)

    def fork(self) -> Fork:
        return self.__base_node.fork()

    def join(self) -> Join:
        return self.__base_node.join()

    def condition(self) -> Condition:
        return self.__base_node.condition()

    def generate(self) -> str:
        return self.__generator.generate_graph_diagram()

    def __repr__(self):
        return f"'{self.title}', {self.generator_cls.__name__}"  # pragma: nocover

    def __str__(self):
        return self.generate()



import typing

from umlcharter.charts.sequence_diagram import (
    SequenceDiagram,
    SequenceDiagramParticipant,
    SequenceDiagramParticipantGroup,
    Step,
    ParticipantActivationControl,
    ForwardStep,
    GroupControl,
    LoopControl,
    ConditionControl,
    CaseControl,
    ReturnStep,
    NoteStep,
)


class SequenceDiagramOrgSequenceDiagram:
    @staticmethod
    def _line_break(string: str) -> str:
        """Some places allow line break as \n"""
        return string.replace("\n", "\\n")

    @staticmethod
    def _remove_line_breaks(string: str) -> str:
        """Some places do not allow line breaks, replace these with just a plane space"""
        return string.replace("\n", " ")

    @classmethod
    def generate(cls, sequence_diagram: SequenceDiagram) -> str:
        participants: typing.Dict[
            SequenceDiagramParticipantGroup, typing.List[SequenceDiagramParticipant]
        ] = sequence_diagram._SequenceDiagram__participants  # noqa
        sequence: typing.List[Step] = sequence_diagram._SequenceDiagram__sequence  # noqa

        first_case: bool = False
        last_targeted_participant: SequenceDiagramParticipant | None = None
        aliases: typing.Dict[SequenceDiagramParticipant, str] = {}
        aliases_counter: int = 1

        participant_types_map = {
            "default": "participant",
            "actor": "actor",
            "boundary": "boundary",
            "control": "control",
            "entity": "entity",
        }

        generated = f"title {cls._line_break(sequence_diagram.title)}\n"
        for group, group_participants in participants.items():
            if group.title:
                generated += (
                    f"participantgroup{group.color.as_hex() if group.color else ''} "
                    f"**{cls._line_break(group.title)}**\n"
                )

            for participant in group_participants:
                # define initial last targeted participant
                if not last_targeted_participant:
                    last_targeted_participant = participant

                aliases[participant] = f"p{aliases_counter}"
                aliases_counter += 1

                generated += (
                    f'{participant_types_map[participant.type_]} "{cls._line_break(participant.title)}" as '
                    f"{aliases[participant]}"
                )
                generated += (
                    f"{participant.color.as_hex() if participant.color else ''}\n"
                )

            if group.title:
                generated += "end\n"

        for step in sequence:
            if isinstance(step, ParticipantActivationControl):
                if step.is_active:
                    generated += f"activate {aliases[step.participant]}{step.color.as_hex() if step.color else ''}\n"
                else:
                    generated += f"deactivate {aliases[step.participant]}\n"

            if isinstance(step, ForwardStep):
                generated += (
                    f"{aliases[step.from_participant]}->{aliases[step.to_participant]}: "
                    f"{cls._line_break(step.text)}\n"
                )
                last_targeted_participant = step.to_participant

            if isinstance(step, ReturnStep):
                generated += (
                    f"{aliases[step.from_participant]}-->{aliases[step.to_participant]}: "
                    f"{cls._line_break(step.text)}\n"
                )
                last_targeted_participant = step.to_participant

            if isinstance(step, GroupControl):
                if step.is_active:
                    generated += (
                        f"group{step.color.as_hex() if step.color else ''} "
                        f"[{cls._remove_line_breaks(step.text)}]\n"
                    )
                else:
                    generated += "end\n"

            if isinstance(step, LoopControl):
                if step.is_active:
                    generated += (
                        f"loop{step.color.as_hex() if step.color else ''} "
                        f"{cls._remove_line_breaks(step.how_many_iterations)}\n"
                    )
                else:
                    generated += "end\n"

            if isinstance(step, ConditionControl):
                if step.is_active:
                    generated += f"alt{step.color.as_hex() if step.color else ''}"
                    first_case = True
                else:
                    generated += "end\n"
                    first_case = False

            if isinstance(step, CaseControl):
                if step.is_active:
                    if first_case:
                        generated += f" {cls._remove_line_breaks(step.text)}\n"
                        first_case = False
                    else:
                        generated += f"else {cls._remove_line_breaks(step.text)}\n"

            if isinstance(step, NoteStep):
                generated += (
                    f"note right of {aliases[last_targeted_participant]}{step.color.as_hex() if step.color else ''}: "
                    f"{cls._line_break(step.text)}\n"
                )

        return generated



import typing
from collections import Counter
from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import chain

from umlcharter.charts.common import BaseChart, ChartingException, Colored
from umlcharter.generators.base import IChartGenerator


class Step:
    pass


@dataclass
class Control(Step):
    is_active: bool


@dataclass
class LoopControl(Colored, Control):
    how_many_iterations: typing.Union[str, None] = None


@dataclass
class GroupControl(Colored, Control):
    text: typing.Union[str, None] = None


@dataclass
class CaseControl(Colored, Control):
    text: typing.Union[str, None] = None


@dataclass
class ConditionControl(Colored, Control):
    pass


@dataclass
class ParticipantActivationControl(Colored, Control):
    participant: "SequenceDiagramParticipant"


@dataclass
class ForwardStep(Step):
    text: str
    from_participant: "SequenceDiagramParticipant"
    to_participant: "SequenceDiagramParticipant"


@dataclass
class ReturnStep(Step):
    text: str
    from_participant: "SequenceDiagramParticipant"
    to_participant: "SequenceDiagramParticipant"


@dataclass
class NoteStep(Colored, Step):
    text: str


@dataclass
class SequenceDiagramParticipant(Colored):
    sequence_ref: "SequenceDiagram"
    title: str
    type_: typing.Literal["actor", "boundary", "control", "entity", "default"] = field(
        init=False, default="default"
    )

    def __check_can_set_type(self):
        if self.type_ != "default":
            raise ChartingException(
                f"The type of the participant must be set only once. The current type is '{self.type_}'"
            )

    def as_actor(self):
        self.__check_can_set_type()
        self.type_ = "actor"
        return self

    def as_boundary(self):
        self.__check_can_set_type()
        self.type_ = "boundary"
        return self

    def as_control(self):
        self.__check_can_set_type()
        self.type_ = "control"
        return self

    def as_entity(self):
        self.__check_can_set_type()
        self.type_ = "entity"
        return self

    def __check_if_interaction_is_possible(self, to: "SequenceDiagramParticipant"):
        """
        The check is based on the purpose of the participant, according to the ECB.
        https://en.wikipedia.org/wiki/Entity-control-boundary

        If the type of the participant is not "default" then only the specific interactions are available:

        - Actors may only know and communicate with boundaries.
        - Boundaries may communicate with actors and controls only.
        - Controls may know and communicate with boundaries and entities, and if needed other controls.
        - Entities may only know about other entities but could communicate also with controls.

        |          | default | actor | boundary | control | entity |
        ------------------------------------------------------------
        | default  |    v    |   v   |     v    |    v    |   v    |
        | actor    |    v    |       |     v    |         |        |
        | boundary |    v    |   v   |          |    v    |        |
        | control  |    v    |       |     v    |    v    |   v    |
        | entity   |    v    |       |          |    v    |   v    |
        """
        allowed = (
            {"default", "default"},
            {"default", "actor"},
            {"default", "boundary"},
            {"default", "control"},
            {"default", "entity"},
            {"actor", "boundary"},
            {"boundary", "control"},
            {"control", "control"},
            {"control", "entity"},
            {"entity", "entity"},
        )

        if {self.type_, to.type_} not in allowed:
            raise ChartingException(
                f"The interaction between '{self.type_}' to '{to.type_}' is not allowed. "
                f"Please correct the types of the participants or remove the use of the types "
                f"if you do not really care about it."
            )

    def __add_step(
        self, step: typing.Union[ForwardStep, ReturnStep, ParticipantActivationControl]
    ):
        self.sequence_ref._SequenceDiagram__add_step(step)  # noqa

    def go_to(
        self, to: "SequenceDiagramParticipant", text: str = ""
    ) -> "SequenceDiagramParticipant":
        self.__check_if_interaction_is_possible(to)
        self.__add_step(ForwardStep(text, from_participant=self, to_participant=to))
        return to

    def return_to(
        self, to: "SequenceDiagramParticipant", text: str = ""
    ) -> "SequenceDiagramParticipant":
        self.__check_if_interaction_is_possible(to)
        self.__add_step(ReturnStep(text, from_participant=self, to_participant=to))
        return to

    @contextmanager
    def activate(self, color: typing.Optional[str] = None):
        """
        Explicitly activate the participant in the diagram on start of the context manager and ends on its exit.
        Can be used to activate the participant *before* it will participate in any activity
        """
        self.__add_step(
            ParticipantActivationControl(is_active=True, participant=self, _color=color)
        )
        yield None
        self.__add_step(
            ParticipantActivationControl(
                is_active=False, participant=self, _color=color
            )
        )

    def __hash__(self):
        return hash(self.title)

    def __repr__(self):
        return f"Participant ({self.title})"  # pragma: nocover


@dataclass
class SequenceDiagramParticipantGroup(Colored):
    """
    Container used to visually join (group, encompass, box) participants.
    """

    title: typing.Optional[str]

    def __hash__(self):
        return hash(self.title)


@dataclass
class SequenceDiagram(BaseChart):
    """
    A sequence of the steps that must be rendered to the diagram DSL - depending on the chosen renderer.

    :title: The title of the diagram to display on the top of the diagram
    :generator_cls: the class of the generator to be used to generate the diagram
    :auto_activation: The flag used to track whether the participant should be activated every time it
        has evoked the action to another participant.
        Once the control flow has returned back and the initial active participant was the target of the action, the
        active participant must be deactivated.
        True by default.
    """

    title: str
    generator_cls: typing.Type[IChartGenerator]
    auto_activation: bool = True

    __participants: typing.Dict[
        SequenceDiagramParticipantGroup, typing.List[SequenceDiagramParticipant]
    ] = field(init=False)
    __sequence: typing.List[Step] = field(init=False)
    __auto_activation_stack: typing.List[
        typing.Union[
            typing.Tuple[SequenceDiagramParticipant, SequenceDiagramParticipant],
            typing.Tuple[None, SequenceDiagramParticipant],
        ]
    ] = field(init=False)
    __generator: IChartGenerator = field(init=False)
    __inside_condition: bool = field(init=False)
    __default_group: SequenceDiagramParticipantGroup = field(init=False)

    def __post_init__(self):
        self.__default_group = SequenceDiagramParticipantGroup(title=None, _color=None)
        self.__participants = {self.__default_group: []}
        self.__sequence = []
        self.__inside_condition = False
        self.__auto_activation_stack = []
        self.__generator = self.generator_cls(self)

    def participant(
        self, title: str, color: typing.Optional[str] = None
    ) -> SequenceDiagramParticipant:
        # NB: every participant must have a unique name
        all_registered_participant_titles = [
            _.title for _ in chain.from_iterable(self.__participants.values())
        ]
        if title in all_registered_participant_titles:
            raise ChartingException(
                f"Sequence diagram already contains participant {title}. "
                f"All participants must have unique titles."
            )
        participant = SequenceDiagramParticipant(
            title=title, sequence_ref=self, _color=color
        )
        # add the participant to the default group
        if self.__default_group not in self.__participants:
            self.__participants[self.__default_group] = []
        self.__participants[self.__default_group].append(participant)
        return participant

    def group_participants(
        self,
        title: str,
        *participants: SequenceDiagramParticipant,
        color: typing.Optional[str] = None,
    ) -> None:
        """
        Visually join (group, encompass, box) participants to emphasize the connections or similarities between them.

        NB 1: the name of this new group must be unique.
        NB 2: every participant can participate in one group only
        """
        all_registered_groups = [_.title for _ in self.__participants]
        if not title or title in all_registered_groups:
            raise ChartingException(
                "The given name of the named group of joint participants "
                "must be unique and not empty."
            )

        for participant in participants:
            if participant not in self.__participants.get(self.__default_group, []):
                raise ChartingException(
                    f"The participant {participant.title} does not belong to the default group "
                    "and therefor cannot be moved to the named one."
                )

        # remove the mentioned participants from the default group and put them to the new one
        self.__participants[self.__default_group] = [
            participant
            for participant in self.__participants.get(self.__default_group, [])
            if participant not in participants
        ]
        # there must be no empty groups
        if not self.__participants.get(self.__default_group):
            self.__participants.pop(self.__default_group, None)

        self.__participants[
            SequenceDiagramParticipantGroup(title=title, _color=color)
        ] = list(participants)

    def note(self, text: str, color: typing.Optional[str] = None) -> None:
        """
        Add the note plate with the given text somewhere inside the diagram
        """
        self.__add_step(NoteStep(text=text, _color=color))

    @contextmanager
    def loop(
        self, how_many_iterations: str, color: typing.Optional[str] = None
    ) -> None:
        """
        Explicitly mark the following sequence of steps as performed in the loop
        """
        self.__add_step(
            LoopControl(
                is_active=True, how_many_iterations=how_many_iterations, _color=color
            )
        )
        yield None
        self.__add_step(LoopControl(is_active=False, _color=color))

    def return_(self, text: str = ""):
        if not self.auto_activation:
            raise ChartingException(
                "The method .return_() can be used only when diagram have been initialized"
                "with `auto_activation=True`. "
                "Please initialize the diagram with `auto_activation=True` "
                "or explicitly use .return_to() for the participant objects."
            )
        try:
            previously_active_participant = self.__auto_activation_stack[-1]
        except IndexError:
            raise ChartingException(
                "Sequence diagram stack does not hold the previous participant to return to."
            )
        self.__add_step(
            ReturnStep(
                text,
                from_participant=previously_active_participant[1],
                to_participant=previously_active_participant[0],
            )
        )

    @contextmanager
    def group(self, text: str, color: typing.Optional[str] = None) -> None:
        """
        Explicitly mark the following sequence of steps as performed in the group
        """
        self.__add_step(GroupControl(is_active=True, text=text, _color=color))
        yield None
        self.__add_step(GroupControl(is_active=False, _color=color))

    @contextmanager
    def condition(self, color: typing.Optional[str] = None) -> None:
        """
        Explicitly mark the following sequence of steps as performed within some logical "if - else" block
        """
        self.__add_step(ConditionControl(is_active=True, _color=color))
        self.__inside_condition = True
        yield
        self.__add_step(ConditionControl(is_active=False, _color=color))
        self.__inside_condition = False

    @contextmanager
    def case(self, text: str, color: typing.Optional[str] = None):
        """
        Explicitly mark the following sequence of steps as performed within the specific condition
        """
        self.__add_step(CaseControl(is_active=True, text=text, _color=color))
        yield
        self.__add_step(CaseControl(is_active=False, _color=color))

    def __active_participants(self) -> typing.List[SequenceDiagramParticipant]:
        """
        If the number of activation `ParticipantActivationControl` associated with the registered participant
        is above of the number of deactivation ones, then the participant is active
        """
        c = Counter()
        for step in self.__sequence:
            if isinstance(step, ParticipantActivationControl):
                c[step.participant] += 1 if step.is_active else -1
        return list(+c)

    def __add_step(self, step: Step):
        if self.__inside_condition:
            # explicitly require the "CaseControl" to always happen right after the "ConditionControl"
            previous_step = self.__sequence[-1]
            if isinstance(previous_step, ConditionControl) and previous_step.is_active:
                if not isinstance(step, CaseControl):
                    raise ChartingException(
                        "After `with .condition()` the next step must be always `with .case()` "
                        "with the definition of the condition. "
                        "Please check the examples from the project repo."
                    )
        else:
            # do not allow "CaseControl" being used outside the condition
            if isinstance(step, CaseControl):
                raise ChartingException(
                    "Context manager `with .case()` cannot be used separately outside of the "
                    "`with .condition()` context manager. "
                    "Please check the examples from the project repo."
                )

        if self.auto_activation:
            # If auto_activation is enabled,
            # every time we add a regular step transferring the control to another participant,
            # the activation must be started.
            # And every time the flow returns to the previously activated participant, its activation must be ended.

            if isinstance(step, ForwardStep):
                if step.to_participant is step.from_participant:
                    # Self-targeting is a special case for the auto-activated sequence diagram, because it must activate
                    # self and deactivate self right after the call.
                    # Also, for simplicity of auto-activation interpretation, the participant must not be activated if
                    # it is already activated
                    if step.to_participant in self.__active_participants():
                        self.__sequence.append(step)
                    else:
                        with step.to_participant.activate():
                            self.__sequence.append(step)
                    return

                if not self.__auto_activation_stack:
                    # If stack is empty, the very first participant starting the flow must be activated as well.
                    self.__auto_activation_stack.append((None, step.from_participant))
                    self.__sequence.append(
                        ParticipantActivationControl(
                            is_active=True,
                            participant=step.from_participant,
                            _color=None,
                        )
                    )

                self.__sequence.append(step)

                if (
                    self.__auto_activation_stack
                    and step.to_participant != self.__auto_activation_stack[-1][-1]
                ):
                    # If the flow has been passed to the participant that is not currently considered as active,
                    # then activate it.
                    self.__auto_activation_stack.append(
                        (step.from_participant, step.to_participant)
                    )
                    self.__sequence.append(
                        ParticipantActivationControl(
                            is_active=True, participant=step.to_participant, _color=None
                        )
                    )

            elif isinstance(step, ReturnStep):
                self.__sequence.append(step)

                if (
                    self.__auto_activation_stack
                    and (step.to_participant, step.from_participant)
                    == self.__auto_activation_stack[-1]
                ):
                    # If we are passing the flow back exactly to the participant,
                    # that previously has passed the control to us -
                    # deactivate the current participant.
                    self.__auto_activation_stack.pop()
                    self.__sequence.append(
                        ParticipantActivationControl(
                            is_active=False,
                            participant=step.from_participant,
                            _color=None,
                        )
                    )

                if self.__auto_activation_stack == [(None, step.to_participant)]:
                    # If we have returned back to the very first participant
                    # that has started the stack of the calls, then also deactivate it.
                    self.__auto_activation_stack.pop()
                    self.__sequence.append(
                        ParticipantActivationControl(
                            is_active=False,
                            participant=step.to_participant,
                            _color=None,
                        )
                    )
            else:
                self.__sequence.append(step)
        else:
            self.__sequence.append(step)

    def generate(self) -> str:
        return self.__generator.generate_sequence_diagram()

    def __repr__(self):
        return f"'{self.title}', {self.generator_cls.__name__}"  # pragma: nocover

    def __str__(self):
        return self.generate()



from unittest.mock import Mock

import pytest

from umlcharter import GraphDiagram, Mermaid
from umlcharter.charts.common import ChartingException


class TestGraphDiagram:
    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram No Nodes
---
stateDiagram-v2
""",
            ),
        ),
    )
    def test_no_nodes(self, generator_cls, output):
        gd = GraphDiagram("Diagram\nNo Nodes", generator_cls=generator_cls)
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram Only Nodes
---
stateDiagram-v2
state "Node #1" as n2
state "Node #2" as n3
""",
            ),
        ),
    )
    def test_only_nodes(self, generator_cls, output):
        gd = GraphDiagram("Diagram Only Nodes", generator_cls=generator_cls)
        gd.node("Node #1")
        gd.node("Node #2")
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram Nodes & Routes
---
stateDiagram-v2
state "Node #1" as n2
state "Node #2" as n3
n2 --> n3 : Hello
n3 --> n2 : Hi!
""",
            ),
        ),
    )
    def test_nodes_with_routes(self, generator_cls, output):
        gd = GraphDiagram("Diagram Nodes & Routes", generator_cls=generator_cls)
        n1 = gd.node("Node #1")
        n2 = gd.node("Node #2")
        n1.go_to(n2, "Hello")
        n2.go_to(n1, "Hi!")
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram Nodes & Routes
---
stateDiagram-v2
state "Node #1" as n2
state "Node #2" as n3
[*] --> n2 : Start!
n2 --> n3 : Hello!
n3 --> n2 : Hi!
n3 --> [*] : Finish!
""",
            ),
        ),
    )
    def test_start_and_finish(self, generator_cls, output):
        gd = GraphDiagram("Diagram Nodes & Routes", generator_cls=generator_cls)
        n1 = gd.node("Node #1")
        n2 = gd.node("Node #2")
        start = gd.start
        finish = gd.finish
        start.go_to(n1, "Start!")
        n1.go_to(n2, "Hello!")
        n2.go_to(n1, "Hi!")
        n2.go_to(finish, "Finish!")
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram Nested Groups
---
stateDiagram-v2
state "Group #1" as n2 {
  state "Nested Group #1" as n5 {
    state "Nested Node #1" as n8
  }
  [*] --> n5 : Go deeper!
  n5 --> [*] : It was deep, indeed
}
classDef cd_n2 fill:#769D8F
class n2 cd_n2
state "Group #2" as n9
n2 --> n9 : Inter-group route
""",
            ),
        ),
    )
    def test_nested_nodes(self, generator_cls, output):
        gd = GraphDiagram("Diagram Nested Groups", generator_cls=generator_cls)
        green_color = "769D8F"
        group1 = gd.node("Group #1", color=green_color)
        group2 = gd.node("Group #2")
        nested_group1 = group1.node("Nested Group #1", green_color)
        nested_group1.node("Nested Node #1", green_color)
        group1.go_to(group2, "Inter-group route")
        group1.start.go_to(nested_group1, "Go deeper!")
        nested_group1.go_to(group1.finish, "It was deep, indeed")
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram Joins & Forks
---
stateDiagram-v2
state "Short Path" as n2
classDef cd_n2 fill:#769D8F
class n2 cd_n2
state "Long Path #1" as n3
state "Long Path #2" as n4
state "Long Path #3" as n5
classDef cd_n5 fill:#769D8F
class n5 cd_n5
state n6 <<fork>>
state n7 <<join>>
[*] --> n6 : Choose your path
n2 --> n7 : It was easy, ha!
n3 --> n4 : 
n4 --> n5 : ... still going ...
n5 --> n7 : 
n6 --> n2 : Choose the easy one
n6 --> n3 : Choose the hard one
n7 --> [*] : 
""",
            ),
        ),
    )
    def test_join_and_fork(self, generator_cls, output):
        gd = GraphDiagram("Diagram Joins & Forks", generator_cls=generator_cls)
        green_color = "769D8F"
        n1 = gd.node("Short Path", green_color)
        n2 = gd.node("Long Path #1")
        n3 = gd.node("Long Path #2")
        n4 = gd.node("Long Path #3", color=green_color)
        fork = gd.fork()
        gd.start.go_to(fork, "Choose your path")
        fork.go_to(n1, "Choose the easy one")
        fork.go_to(n2, "Choose the hard one")
        n2.go_to(n3)
        n3.go_to(n4, "... still going ...")
        join = gd.join()
        n4.go_to(join)
        n1.go_to(join, "It was easy, ha!")
        join.go_to(gd.finish)
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Diagram Conditions
---
stateDiagram-v2
state "Node #1" as n2
state n3 <<choice>>
state "Node (conditional) #2" as n4
[*] --> n2 : Go to first
n2 --> n3 : Make a choice
n3 --> n4 : Need a second one
n3 --> [*] : Nah, try again
n4 --> [*] : 
""",
            ),
        ),
    )
    def test_condition(self, generator_cls, output):
        gd = GraphDiagram("Diagram Conditions", generator_cls=generator_cls)
        n1 = gd.node("Node #1")
        c1 = gd.condition()
        n2 = gd.node("Node (conditional) #2")
        gd.start.go_to(n1, "Go to first")
        n1.go_to(c1, "Make a choice")
        c1.go_to(n2, "Need a second one")
        n2.go_to(gd.finish)
        c1.go_to(gd.finish, "Nah, try again")
        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """---
title: Complex Diagram With Notes
---
stateDiagram-v2
state "Group" as n2 {
  state "Nested Node" as n5
  note right of n5
Note for the nested node
  end note
}
note right of n2
Note for the group
end note
state "Outer Node" as n6
note right of n6
It is possible to have multiple notes...
end note
note right of n6
...and also split them
in multi lines
end note
state n7 <<fork>>
note right of n7
Note for fork
end note
state n8 <<choice>>
note right of n8
Note for condition
end note
state n9 <<join>>
note right of n9
Note for join
end note
n2 --> n7 : 
n7 --> n8 : 
n7 --> n9 : 
n8 --> n9 : 
n8 --> [*] : 
""",
            ),
        ),
    )
    def test_complex_with_notes(self, generator_cls, output):
        gd = GraphDiagram("Complex Diagram With Notes", generator_cls=generator_cls)
        group = gd.node("Group")
        nested_node = group.node("Nested Node")
        node = gd.node("Outer Node")
        fork = gd.fork()
        condition = gd.condition()
        join = gd.join()

        group.go_to(fork)
        fork.go_to(condition)
        condition.go_to(join)
        condition.go_to(gd.finish)
        fork.go_to(join)

        group.note("Note for the group")
        nested_node.note("Note for the nested node")
        node.note("It is possible to have multiple notes...")
        node.note("...and also split them\nin multi lines")

        fork.note("Note for fork")
        condition.note("Note for condition")
        join.note("Note for join")

        assert str(gd) == output

    @pytest.mark.parametrize(
        "generator_cls,is_vertical,output",
        (
            (
                Mermaid,
                True,
                """---
title: Diagram Different Orientation
---
stateDiagram-v2
state "Node: #1" as n2
state "Node: #2" as n3
state "Node: #3" as n4
state "Node: #4" as n5
n2 --> n3 : 
n3 --> n4 : 
n4 --> n5 : 
n5 --> n2 : We are rolling yay!
""",
            ),
            (
                Mermaid,
                False,
                """---
title: Diagram Different Orientation
---
stateDiagram-v2
direction LR
state "Node: #1" as n2
state "Node: #2" as n3
state "Node: #3" as n4
state "Node: #4" as n5
n2 --> n3 : 
n3 --> n4 : 
n4 --> n5 : 
n5 --> n2 : We are rolling yay!
""",
            ),
        ),
    )
    def test_orientation(self, generator_cls, is_vertical, output):
        gd = GraphDiagram(
            "Diagram: Different Orientation",
            generator_cls=generator_cls,
            is_vertical=is_vertical,
        )
        n1 = gd.node("Node: #1")
        n2 = gd.node("Node: #2")
        n1.go_to(n2)
        n3 = gd.node("Node: #3")
        n2.go_to(n3)
        n4 = gd.node("Node: #4")
        n3.go_to(n4)
        n4.go_to(n1, "We are rolling: yay!")
        assert str(gd) == output

    def test_invalid_color_string(self):
        gd = GraphDiagram("Invalid color", Mock)
        with pytest.raises(ChartingException):
            gd.node("Invalid color", "red")

    def test_same_titles_on_the_same_level(self):
        gd = GraphDiagram("Same titles", Mock)
        gd.node("Title")
        with pytest.raises(ChartingException):
            gd.node("Title")

    def test_cannot_interact_between_levels(self):
        gd = GraphDiagram("Different levels", Mock)
        outer_node = gd.node("Outer")
        nested_node = gd.node("Group").node("Nested")
        with pytest.raises(ChartingException):
            outer_node.go_to(nested_node)

    def test_cannot_build_same_graph_edge_twice(self):
        gd = GraphDiagram("Duplicated graph edges", Mock)
        n1 = gd.node("Node #1")
        n2 = gd.node("Node #2")
        n1.go_to(n2)
        with pytest.raises(ChartingException):
            n1.go_to(n2)

    def test_cannot_go_to_abstract_start(self):
        gd = GraphDiagram("Start is not a destination", Mock)
        n1 = gd.node("Node #1")
        with pytest.raises(ChartingException):
            n1.go_to(gd.start)

    def test_cannot_go_from_abstract_finish(self):
        gd = GraphDiagram("Finish is not a strtating point", Mock)
        n1 = gd.node("Node #1")
        with pytest.raises(ChartingException):
            gd.finish.go_to(n1)

    def test_cannot_directly_link_start_and_finish(self):
        gd = GraphDiagram("Directly link start and finish", Mock)
        with pytest.raises(ChartingException):
            gd.start.go_to(gd.finish)



from umlcharter.charts.graph_diagram import (
    Node,
    Start,
    Finish,
    Condition,
    Join,
    Fork,
    GraphDiagram,
)


class MermaidGraphDiagram:
    @staticmethod
    def _remove_line_breaks(string: str) -> str:
        """
        Some places do not allow line breaks, replace these with just a plane space.
        Also, these places do not allow ':' symbol
        """
        return string.replace("\n", " ").replace(":", "")

    @classmethod
    def generate(cls, sequence_diagram: GraphDiagram) -> str:
        aliases = {}

        generated = f"---\ntitle: {cls._remove_line_breaks(sequence_diagram.title)}\n---\nstateDiagram-v2\n"
        if not sequence_diagram.is_vertical:
            # default direction is top -> bottom, specify if it is not default
            generated += "direction LR\n"

        def recursive_graph_generation(
            generated_dsl: str, node_to_process: Node, depth: int
        ) -> str:
            ident = " " * depth
            inner_graph = node_to_process._Node__inner_graph  # noqa
            # iterate once to define the states first...
            for node, routes in inner_graph.items():
                if isinstance(node, Start) or isinstance(node, Finish):
                    aliases[node] = "[*]"
                else:
                    node_alias = f"n{len(aliases)}"
                    aliases[node] = node_alias
                    if isinstance(node, Condition):
                        generated_dsl += f"{ident}state {node_alias} <<choice>>\n"
                    if isinstance(node, Join):
                        generated_dsl += f"{ident}state {node_alias} <<join>>\n"
                    if isinstance(node, Fork):
                        generated_dsl += f"{ident}state {node_alias} <<fork>>\n"
                    if isinstance(node, Node):
                        if node.is_group():
                            generated_dsl += (
                                f'{ident}state "{node.text}" as {node_alias} '
                                f'{{\n{recursive_graph_generation("", node, depth + 2)}{ident}}}\n'
                            )
                        else:
                            generated_dsl += (
                                f'{ident}state "{node.text}" as {node_alias}\n'
                            )
                        if node_to_process.is_top_level() and node.color:
                            # NB: mermaid does not support styling for the nodes inside composite states ("groups") yet.
                            # So the styling will be applied ONLY to the nodes on the most top level of the graph
                            class_def = f"cd_{node_alias}"
                            generated_dsl += (
                                f"{ident}classDef {class_def} fill:{node.color.as_hex()}\n"
                                f"{ident}class {node_alias} {class_def}\n"
                            )
                notes: list[str] = getattr(node, "_notes", [])
                for note in notes:
                    generated_dsl += (
                        f"{ident}note right of {node_alias}\n{note}\n{ident}end note\n"
                    )

            # ...second run is to define the routes between the nodes
            for node, routes in inner_graph.items():
                for route in routes:
                    to_node, route_text = route
                    generated_dsl += f"{ident}{aliases[node]} --> {aliases[to_node]} : {cls._remove_line_breaks(route_text)}\n"

            return generated_dsl

        base_node: Node = sequence_diagram._GraphDiagram__base_node  # noqa
        return recursive_graph_generation(generated, base_node, 0)



import setuptools
import re

with open("README.md", "rt") as f:
    long_description = f.read()

with open("umlcharter/__init__.py", "rt") as f:
    version = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)

setuptools.setup(
    name="umlcharter",
    version=version,
    author="Mikalai Yurkin",
    author_email="yurkin.mikalai@gmail.com",
    description="Python package for quick diagrams creation without a need to learn new DSL",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mikalaiyurkin/charter",
    packages=setuptools.find_packages(exclude=("tests",)),
    extras_require={"dev": ["pytest", "pytest-cov", "pre-commit"]},
    python_requires=">=3.9",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
    ],
)



from unittest.mock import Mock

import pytest

from umlcharter import SequenceDiagram, Mermaid, PlantUML, D2, SequenceDiagramOrg
from umlcharter.charts.common import ChartingException


class TestSequenceDiagram:
    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Empty
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Empty
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Empty {
shape: sequence_diagram
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Empty
""",
            ),
        ),
    )
    def test_no_participants(self, generator_cls, output):
        sd = SequenceDiagram("Diagram Empty", generator_cls=generator_cls)
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Only Participants
participant p1 as First
participant p2 as Second
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Only Participants
participant "First" as p1 
participant "Second" as p2 
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Only Participants {
shape: sequence_diagram
p1: First 
p2: Second 
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Only Participants
participant "First" as p1
participant "Second" as p2
""",
            ),
        ),
    )
    def test_only_participants(self, generator_cls, output):
        sd = SequenceDiagram("Diagram Only Participants", generator_cls=generator_cls)
        sd.participant("First")
        sd.participant("Second")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,auto_activation,output",
        (
            (
                Mermaid,
                True,
                """sequenceDiagram
Title: Diagram Interaction and Auto Activation
participant p1 as First<br/>Participant
participant p2 as Second<br/>Participant
participant p3 as Third<br/>Participant
participant p4 as Fourth<br/>Participant
activate p1
p1->>p2: Go to second
activate p2
p2-->>p1: Return to first
deactivate p2
deactivate p1
activate p1
p1->>p3: Go to third
activate p3
p3->>p4: Go to fourth
activate p4
p4->>p4: Go to self
p4-->>p3: Return to third
deactivate p4
p3-->>p1: Return to first
deactivate p3
deactivate p1
""",
            ),
            (
                Mermaid,
                False,
                """sequenceDiagram
Title: Diagram Interaction and Auto Activation
participant p1 as First<br/>Participant
participant p2 as Second<br/>Participant
participant p3 as Third<br/>Participant
participant p4 as Fourth<br/>Participant
p1->>p2: Go to second
p2-->>p1: Return to first
p1->>p3: Go to third
p3->>p4: Go to fourth
p4->>p4: Go to self
p4-->>p3: Return to third
p3-->>p1: Return to first
""",
            ),
            (
                PlantUML,
                True,
                """@startuml
title: Diagram Interaction\\nand Auto Activation
participant "First\\nParticipant" as p1 
participant "Second\\nParticipant" as p2 
participant "Third\\nParticipant" as p3 
participant "Fourth\\nParticipant" as p4 
activate p1 
p1->p2: Go to second
activate p2 
p2-->p1: Return to first
deactivate p2
deactivate p1
p1 -[hidden]-> p1
activate p1 
p1->p3: Go to third
activate p3 
p3->p4: Go to fourth
activate p4 
p4->p4: Go to self
p4-->p3: Return to third
deactivate p4
p3-->p1: Return to first
deactivate p3
deactivate p1
@enduml
""",
            ),
            (
                PlantUML,
                False,
                """@startuml
title: Diagram Interaction\\nand Auto Activation
participant "First\\nParticipant" as p1 
participant "Second\\nParticipant" as p2 
participant "Third\\nParticipant" as p3 
participant "Fourth\\nParticipant" as p4 
p1->p2: Go to second
p2-->p1: Return to first
p1->p3: Go to third
p3->p4: Go to fourth
p4->p4: Go to self
p4-->p3: Return to third
p3-->p1: Return to first
@enduml
""",
            ),
            (
                D2,
                True,
                """title: Diagram Interaction\\nand Auto Activation {
shape: sequence_diagram
p1: First\\nParticipant 
p2: Second\\nParticipant 
p3: Third\\nParticipant 
p4: Fourth\\nParticipant 
p1.0 -> p2.1: Go to second
p2.1 -> p1.0: Return to first {style.stroke-dash: 3}
p1.2 -> p3.3: Go to third
p3.3 -> p4.4: Go to fourth
p4.4 -> p4.4: Go to self
p4.4 -> p3.3: Return to third {style.stroke-dash: 3}
p3.3 -> p1.2: Return to first {style.stroke-dash: 3}
}
""",
            ),
            (
                D2,
                False,
                """title: Diagram Interaction\\nand Auto Activation {
shape: sequence_diagram
p1: First\\nParticipant 
p2: Second\\nParticipant 
p3: Third\\nParticipant 
p4: Fourth\\nParticipant 
p1 -> p2: Go to second
p2 -> p1: Return to first {style.stroke-dash: 3}
p1 -> p3: Go to third
p3 -> p4: Go to fourth
p4 -> p4: Go to self
p4 -> p3: Return to third {style.stroke-dash: 3}
p3 -> p1: Return to first {style.stroke-dash: 3}
}
""",
            ),
            (
                SequenceDiagramOrg,
                True,
                """title Diagram Interaction\\nand Auto Activation
participant "First\\nParticipant" as p1
participant "Second\\nParticipant" as p2
participant "Third\\nParticipant" as p3
participant "Fourth\\nParticipant" as p4
activate p1
p1->p2: Go to second
activate p2
p2-->p1: Return to first
deactivate p2
deactivate p1
activate p1
p1->p3: Go to third
activate p3
p3->p4: Go to fourth
activate p4
p4->p4: Go to self
p4-->p3: Return to third
deactivate p4
p3-->p1: Return to first
deactivate p3
deactivate p1
""",
            ),
            (
                SequenceDiagramOrg,
                False,
                """title Diagram Interaction\\nand Auto Activation
participant "First\\nParticipant" as p1
participant "Second\\nParticipant" as p2
participant "Third\\nParticipant" as p3
participant "Fourth\\nParticipant" as p4
p1->p2: Go to second
p2-->p1: Return to first
p1->p3: Go to third
p3->p4: Go to fourth
p4->p4: Go to self
p4-->p3: Return to third
p3-->p1: Return to first
""",
            ),
        ),
    )
    def test_simple_interaction_and_auto_activation(
        self, generator_cls, auto_activation, output
    ):
        sd = SequenceDiagram(
            "Diagram Interaction\nand Auto Activation",
            generator_cls=generator_cls,
            auto_activation=auto_activation,
        )
        first = sd.participant("First\nParticipant")
        second = sd.participant("Second\nParticipant")
        third = sd.participant("Third\nParticipant")
        fourth = sd.participant("Fourth\nParticipant")

        first.go_to(second, "Go to second").return_to(first, "Return to first")

        first.go_to(third, "Go to third").go_to(fourth, "Go to fourth").go_to(
            fourth, "Go to self"
        ).return_to(third, "Return to third").return_to(first, "Return to first")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Interaction and Manual Activation
participant p1 as First
participant p2 as Second
activate p1
p1->>p2: Go to second
activate p2
p2->>p2: Go to self
p2-->>p1: Return to first
deactivate p2
deactivate p1
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Interaction and Manual Activation
participant "First" as p1 
participant "Second" as p2 #769D8F
activate p1 
p1->p2: Go to second
activate p2 #769D8F
p2->p2: Go to self
p2-->p1: Return to first
deactivate p2
deactivate p1
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Interaction and Manual Activation {
shape: sequence_diagram
p1: First 
p2: Second {
style: {fill: "#769D8F" 
stroke:"#769D8F" }
}
p1.0 -> p2.1: Go to second
p2.1 -> p2.1: Go to self
p2.1 -> p1.0: Return to first {style.stroke-dash: 3}
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Interaction and Manual Activation
participant "First" as p1
participant "Second" as p2#769D8F
activate p1
p1->p2: Go to second
activate p2#769D8F
p2->p2: Go to self
p2-->p1: Return to first
deactivate p2
deactivate p1
""",
            ),
        ),
    )
    def test_simple_interaction_and_manual_activation(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Interaction and Manual Activation",
            generator_cls=generator_cls,
            auto_activation=False,
        )
        green_color = "769D8F"  # some shade of green color
        first = sd.participant("First")
        second = sd.participant("Second", color=green_color)

        with first.activate():
            first.go_to(second, "Go to second")
            with second.activate(color=green_color):
                second.go_to(second, "Go to self")
                second.return_to(first, "Return to first")

        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Self-targeting
participant p1 as A
participant p2 as B
activate p1
p1->>p1: Self-targeting at the beginning
deactivate p1
activate p1
p1->>p2: 
activate p2
p1->>p1: Self-targeting while flow under control of other participant
p2->>p2: Self-targeting while being active
p2-->>p1: 
deactivate p2
deactivate p1
activate p1
p1->>p1: Self-targeting at the end
deactivate p1
""",
            ),
            (
                PlantUML,
                """@startuml
title: Self-targeting
participant "A" as p1 
participant "B" as p2 
activate p1 
p1->p1: Self-targeting at the beginning
deactivate p1
p1 -[hidden]-> p1
activate p1 
p1->p2: 
activate p2 
p1->p1: Self-targeting while flow under control of other participant
p2->p2: Self-targeting while being active
p2-->p1: 
deactivate p2
deactivate p1
p1 -[hidden]-> p1
activate p1 
p1->p1: Self-targeting at the end
deactivate p1
@enduml
""",
            ),
            (
                D2,
                """title: Self-targeting {
shape: sequence_diagram
p1: A 
p2: B 
p1.0 -> p1.0: Self-targeting at the beginning
p1.1 -> p2.2: ''
p1.1 -> p1.1: Self-targeting while flow under control of other participant
p2.2 -> p2.2: Self-targeting while being active
p2.2 -> p1.1: '' {style.stroke-dash: 3}
p1.3 -> p1.3: Self-targeting at the end
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Self-targeting
participant "A" as p1
participant "B" as p2
activate p1
p1->p1: Self-targeting at the beginning
deactivate p1
activate p1
p1->p2: 
activate p2
p1->p1: Self-targeting while flow under control of other participant
p2->p2: Self-targeting while being active
p2-->p1: 
deactivate p2
deactivate p1
activate p1
p1->p1: Self-targeting at the end
deactivate p1
""",
            ),
        ),
    )
    def test_self_targeting_with_auto_activation(self, generator_cls, output):
        sd = SequenceDiagram("Self-targeting", generator_cls, auto_activation=True)
        a = sd.participant("A")
        b = sd.participant("B")
        a.go_to(a, "Self-targeting at the beginning")
        a.go_to(b)
        a.go_to(a, "Self-targeting while flow under control of other participant")
        b.go_to(b, "Self-targeting while being active")
        b.return_to(a)
        a.go_to(a, "Self-targeting at the end")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Interaction and Grouping
participant p1 as First
participant p2 as Second
participant p3 as Third
rect rgb(230, 230, 240, 0.5)
note right of p1: Group enclosing everything
activate p1
p1->>p2: Go to second
activate p2
rect rgb(230, 230, 240, 0.5)
note right of p2: Group enclosing interaction<br/>between second and third
p2->>p3: Go to third
activate p3
p3-->>p2: Return to second
deactivate p3
end
p2-->>p1: Return to first
deactivate p2
deactivate p1
end
activate p3
p3->>p1: 
activate p1
p1-->>p3: Short reverse direct connection
deactivate p1
deactivate p3
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Interaction and Grouping
participant "First" as p1 
participant "Second" as p2 
participant "Third" as p3 
group  Group enclosing everything
activate p1 
p1->p2: Go to second
activate p2 
group#769D8F #769D8F Group enclosing interaction\\nbetween second and third
p2->p3: Go to third
activate p3 
p3-->p2: Return to second
deactivate p3
end
p2-->p1: Return to first
deactivate p2
deactivate p1
end
p3 -[hidden]-> p3
activate p3 
p3->p1: 
activate p1 
p1-->p3: Short reverse direct connection
deactivate p1
deactivate p3
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Interaction and Grouping {
shape: sequence_diagram
p1: First 
p2: Second 
p3: Third 
group1: \\[GROUP\\] Group enclosing everything: {
p1.0 -> p2.1: Go to second
group2: \\[GROUP\\] Group enclosing interaction\\nbetween second and third: {
style: {
fill: "#769D8F" 
}
p2.1 -> p3.2: Go to third
p3.2 -> p2.1: Return to second {style.stroke-dash: 3}
}
p2.1 -> p1.0: Return to first {style.stroke-dash: 3}
}
p3.3 -> p1.4: ''
p1.4 -> p3.3: Short reverse direct connection {style.stroke-dash: 3}
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Interaction and Grouping
participant "First" as p1
participant "Second" as p2
participant "Third" as p3
group [Group enclosing everything]
activate p1
p1->p2: Go to second
activate p2
group#769D8F [Group enclosing interaction between second and third]
p2->p3: Go to third
activate p3
p3-->p2: Return to second
deactivate p3
end
p2-->p1: Return to first
deactivate p2
deactivate p1
end
activate p3
p3->p1: 
activate p1
p1-->p3: Short reverse direct connection
deactivate p1
deactivate p3
""",
            ),
        ),
    )
    def test_grouping(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Interaction and Grouping", generator_cls=generator_cls
        )

        first = sd.participant("First")
        second = sd.participant("Second")
        third = sd.participant("Third")

        with sd.group("Group enclosing everything"):
            first.go_to(second, "Go to second")
            with sd.group(
                "Group enclosing interaction\nbetween second and third", color="769D8F"
            ):
                second.go_to(third, "Go to third").return_to(second, "Return to second")
            second.return_to(first, "Return to first")

        third.go_to(first).return_to(third, "Short reverse direct connection")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Interaction and Loops
participant p1 as First
participant p2 as Second
loop Infinite loop
activate p1
p1->>p2: Send request to second
activate p2
loop Repeat<br/>until available
p2->>p2: Check internal state
end
p2-->>p1: Return response
deactivate p2
deactivate p1
end
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Interaction and Loops
participant "First" as p1 
participant "Second" as p2 
loop  Infinite loop
activate p1 
p1->p2: Send request to second
activate p2 
loop#769D8F #769D8F Repeat\\nuntil available
p2->p2: Check internal state
end
p2-->p1: Return response
deactivate p2
deactivate p1
end
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Interaction and Loops {
shape: sequence_diagram
p1: First 
p2: Second 
loop1: \\[LOOP\\] Infinite loop: {
p1.0 -> p2.1: Send request to second
loop2: \\[LOOP\\] Repeat\\nuntil available: {
style: {
fill: "#769D8F" 
}
p2.1 -> p2.1: Check internal state
}
p2.1 -> p1.0: Return response {style.stroke-dash: 3}
}
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Interaction and Loops
participant "First" as p1
participant "Second" as p2
loop Infinite loop
activate p1
p1->p2: Send request to second
activate p2
loop#769D8F Repeat until available
p2->p2: Check internal state
end
p2-->p1: Return response
deactivate p2
deactivate p1
end
""",
            ),
        ),
    )
    def test_loop(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Interaction and Loops", generator_cls=generator_cls
        )

        first = sd.participant("First")
        second = sd.participant("Second")

        with sd.loop("Infinite loop"):
            first.go_to(second, "Send request to second")
            with sd.loop("Repeat\nuntil available", color="769D8F"):
                second.go_to(second, "Check internal state")
            sd.return_("Return response")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Interaction and Conditions
participant p1 as Viewer
participant p2 as Drama
participant p3 as Comedy
activate p1
p1->>p1: What would I like to watch today?
deactivate p1
alt Want a drama
activate p1
p1->>p2: Watch drama
activate p2
p2-->>p1: Tears and sadness
deactivate p2
deactivate p1
else Want a comedy
activate p1
p1->>p3: Watch comedy
activate p3
p3-->>p1: Laugh a lot
deactivate p3
deactivate p1
end
activate p1
p1->>p1: What would I like to watch tomorrow?
deactivate p1
alt Want a drama
activate p1
p1->>p2: Watch drama
activate p2
p2-->>p1: Tears and sadness
deactivate p2
deactivate p1
else Want a comedy
activate p1
p1->>p3: Watch comedy
activate p3
p3-->>p1: Laugh a lot
deactivate p3
deactivate p1
end
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Interaction and Conditions
participant "Viewer" as p1 
participant "Drama" as p2 
participant "Comedy" as p3 
activate p1 
p1->p1: What would I like to watch today?
deactivate p1
alt #FFFFFF Want a drama
p1 -[hidden]-> p1
activate p1 
p1->p2: Watch drama
activate p2 
p2-->p1: Tears and sadness
deactivate p2
deactivate p1
else #FFFFFF Want a comedy
p1 -[hidden]-> p1
activate p1 
p1->p3: Watch comedy
activate p3 
p3-->p1: Laugh a lot
deactivate p3
deactivate p1
end
p1 -[hidden]-> p1
activate p1 
p1->p1: What would I like to watch tomorrow?
deactivate p1
alt#b2b200 #FFFFFF Want a drama
p1 -[hidden]-> p1
activate p1 
p1->p2: Watch drama
activate p2 
p2-->p1: Tears and sadness
deactivate p2
deactivate p1
else #66b266 Want a comedy
p1 -[hidden]-> p1
activate p1 
p1->p3: Watch comedy
activate p3 
p3-->p1: Laugh a lot
deactivate p3
deactivate p1
end
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Interaction and Conditions {
shape: sequence_diagram
p1: Viewer 
p2: Drama 
p3: Comedy 
p1.0 -> p1.0: What would I like to watch today?
alt1: \\[ALT\\] {
case2: \\[CASE\\] Want a drama: {
p1.1 -> p2.2: Watch drama
p2.2 -> p1.1: Tears and sadness {style.stroke-dash: 3}
}
case3: \\[CASE\\] Want a comedy: {
p1.3 -> p3.4: Watch comedy
p3.4 -> p1.3: Laugh a lot {style.stroke-dash: 3}
}
}
p1.5 -> p1.5: What would I like to watch tomorrow?
alt4: \\[ALT\\] {
style: {
fill: "#b2b200" 
}
case5: \\[CASE\\] Want a drama: {
p1.6 -> p2.7: Watch drama
p2.7 -> p1.6: Tears and sadness {style.stroke-dash: 3}
}
case6: \\[CASE\\] Want a comedy: {
style: {
fill: "#66b266" 
}
p1.8 -> p3.9: Watch comedy
p3.9 -> p1.8: Laugh a lot {style.stroke-dash: 3}
}
}
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Interaction and Conditions
participant "Viewer" as p1
participant "Drama" as p2
participant "Comedy" as p3
activate p1
p1->p1: What would I like to watch today?
deactivate p1
alt Want a drama
activate p1
p1->p2: Watch drama
activate p2
p2-->p1: Tears and sadness
deactivate p2
deactivate p1
else Want a comedy
activate p1
p1->p3: Watch comedy
activate p3
p3-->p1: Laugh a lot
deactivate p3
deactivate p1
end
activate p1
p1->p1: What would I like to watch tomorrow?
deactivate p1
alt#b2b200 Want a drama
activate p1
p1->p2: Watch drama
activate p2
p2-->p1: Tears and sadness
deactivate p2
deactivate p1
else Want a comedy
activate p1
p1->p3: Watch comedy
activate p3
p3-->p1: Laugh a lot
deactivate p3
deactivate p1
end
""",
            ),
        ),
    )
    def test_condition(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Interaction and Conditions",
            generator_cls=generator_cls,
            auto_activation=False,
        )

        viewer = sd.participant("Viewer")
        drama = sd.participant("Drama")
        comedy = sd.participant("Comedy")

        with viewer.activate():
            viewer.go_to(viewer, "What would I like to watch today?")

        with sd.condition():
            with sd.case("Want a drama"):
                with viewer.activate():
                    viewer.go_to(drama, "Watch drama")
                    with drama.activate():
                        drama.return_to(viewer, "Tears and sadness")
            with sd.case("Want a comedy"):
                with viewer.activate():
                    viewer.go_to(comedy, "Watch comedy")
                    with comedy.activate():
                        comedy.return_to(viewer, "Laugh a lot")

        with viewer.activate():
            viewer.go_to(viewer, "What would I like to watch tomorrow?")

        green_yellow_color = "b2b200"
        green_color = "66b266"

        with sd.condition(color=green_yellow_color):
            with sd.case("Want a drama"):
                with viewer.activate():
                    viewer.go_to(drama, "Watch drama")
                    with drama.activate():
                        drama.return_to(viewer, "Tears and sadness")
            with sd.case("Want a comedy", color=green_color):
                with viewer.activate():
                    viewer.go_to(comedy, "Watch comedy")
                    with comedy.activate():
                        comedy.return_to(viewer, "Laugh a lot")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Interaction and Notes
participant p1 as Batman
participant p2 as Bandit
note right of p1: Batman is throwing<br/>a batarang at the bandit
activate p1
p1->>p2: Pheeeeeeu!
activate p2
note right of p2: Batman has missed!
p2-->>p1: A bad day<br/>for the Gotham :(
deactivate p2
deactivate p1
note right of p1: Batman is sad now
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Interaction and Notes
participant "Batman" as p1 
participant "Bandit" as p2 
note right of p1 : Batman is throwing\\na batarang at the bandit
activate p1 
p1->p2: Pheeeeeeu!
activate p2 
note right of p2 #769D8F: Batman has missed!
p2-->p1: A bad day\\nfor the Gotham :(
deactivate p2
deactivate p1
note right of p1 : Batman is sad now
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Interaction and Notes {
shape: sequence_diagram
p1: Batman 
p2: Bandit 
p1."Batman is throwing\\na batarang at the bandit"
p1.0 -> p2.1: Pheeeeeeu!
p2.1."Batman has missed!"
p2.1 -> p1.0: A bad day\\nfor the Gotham :( {style.stroke-dash: 3}
p1."Batman is sad now"
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Interaction and Notes
participant "Batman" as p1
participant "Bandit" as p2
note right of p1: Batman is throwing\\na batarang at the bandit
activate p1
p1->p2: Pheeeeeeu!
activate p2
note right of p2#769D8F: Batman has missed!
p2-->p1: A bad day\\nfor the Gotham :(
deactivate p2
deactivate p1
note right of p1: Batman is sad now
""",
            ),
        ),
    )
    def test_note(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Interaction and Notes",
            generator_cls,
        )

        batman = sd.participant("Batman")
        bandit = sd.participant("Bandit")

        sd.note("Batman is throwing\na batarang at the bandit")
        batman.go_to(bandit, "Pheeeeeeu!")
        sd.note("Batman has missed!", color="769D8F")
        sd.return_("A bad day\nfor the Gotham :(")
        sd.note("Batman is sad now")
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Participants Grouping
participant p1 as Participant 1
participant p2 as Participant 4
box A first group
participant p3 as Participant 2
participant p4 as Participant 3
end
box A second group
participant p5 as Participant 5
end
activate p1
p1->>p3: Pass a message
activate p3
p3->>p4: Pass a message
activate p4
p4->>p2: Pass a message
activate p2
p2->>p5: Message!
activate p5
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Participants Grouping
participant "Participant 1" as p1 
participant "Participant 4" as p2 
box "A first\\ngroup" 
participant "Participant 2" as p3 
participant "Participant 3" as p4 
end box
box "A second\\ngroup" 
participant "Participant 5" as p5 
end box
activate p1 
p1->p3: Pass a message
activate p3 
p3->p4: Pass a message
activate p4 
p4->p2: Pass a message
activate p2 
p2->p5: Message!
activate p5 
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Participants Grouping {
shape: sequence_diagram
p1: Participant 1 
p2: Participant 4 
p3: Participant 2 
p4: Participant 3 
p5: Participant 5 
p1.0 -> p3.1: Pass a message
p3.1 -> p4.2: Pass a message
p4.2 -> p2.3: Pass a message
p2.3 -> p5.4: Message!
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Participants Grouping
participant "Participant 1" as p1
participant "Participant 4" as p2
participantgroup **A first\\ngroup**
participant "Participant 2" as p3
participant "Participant 3" as p4
end
participantgroup **A second\\ngroup**
participant "Participant 5" as p5
end
activate p1
p1->p3: Pass a message
activate p3
p3->p4: Pass a message
activate p4
p4->p2: Pass a message
activate p2
p2->p5: Message!
activate p5
""",
            ),
        ),
    )
    def test_hectic_participant_grouping(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Participants Grouping",
            generator_cls,
        )
        not_grouped_1 = sd.participant("Participant 1")
        grouped_2 = sd.participant("Participant 2")
        grouped_3 = sd.participant("Participant 3")
        not_grouped_4 = sd.participant("Participant 4")
        grouped_5 = sd.participant("Participant 5")

        not_grouped_1.go_to(grouped_2, "Pass a message").go_to(
            grouped_3, "Pass a message"
        ).go_to(not_grouped_4, "Pass a message").go_to(grouped_5, "Message!")

        sd.group_participants("A first\ngroup", grouped_2, grouped_3)
        sd.group_participants("A second\ngroup", grouped_5)
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Diagram Participants Grouping
box A first group
participant p1 as Participant 1
participant p2 as Participant 2
end
box A second group
participant p3 as Participant 3
end
box A third group
participant p4 as Participant 4
participant p5 as Participant 5
end
activate p1
p1->>p2: Pass a message
activate p2
p2->>p3: Pass a message
activate p3
p3->>p4: Pass a message
activate p4
p4->>p5: Message!
activate p5
""",
            ),
            (
                PlantUML,
                """@startuml
title: Diagram Participants Grouping
box "A first\\ngroup" #b2b200
participant "Participant 1" as p1 #769D8F
participant "Participant 2" as p2 #66b266
end box
box "A second\\ngroup" 
participant "Participant 3" as p3 
end box
box "A third\\ngroup" 
participant "Participant 4" as p4 
participant "Participant 5" as p5 
end box
activate p1 
p1->p2: Pass a message
activate p2 
p2->p3: Pass a message
activate p3 
p3->p4: Pass a message
activate p4 
p4->p5: Message!
activate p5 
@enduml
""",
            ),
            (
                D2,
                """title: Diagram Participants Grouping {
shape: sequence_diagram
p1: Participant 1 {
style: {fill: "#769D8F" 
stroke:"#769D8F" }
}
p2: Participant 2 {
style: {fill: "#66b266" 
stroke:"#66b266" }
}
p3: Participant 3 
p4: Participant 4 
p5: Participant 5 
p1.0 -> p2.1: Pass a message
p2.1 -> p3.2: Pass a message
p3.2 -> p4.3: Pass a message
p4.3 -> p5.4: Message!
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Diagram Participants Grouping
participantgroup#b2b200 **A first\\ngroup**
participant "Participant 1" as p1#769D8F
participant "Participant 2" as p2#66b266
end
participantgroup **A second\\ngroup**
participant "Participant 3" as p3
end
participantgroup **A third\\ngroup**
participant "Participant 4" as p4
participant "Participant 5" as p5
end
activate p1
p1->p2: Pass a message
activate p2
p2->p3: Pass a message
activate p3
p3->p4: Pass a message
activate p4
p4->p5: Message!
activate p5
""",
            ),
        ),
    )
    def test_accurate_participant_grouping(self, generator_cls, output):
        sd = SequenceDiagram(
            "Diagram Participants Grouping",
            generator_cls,
        )
        green_color = "769D8F"
        another_green_color = "66b266"
        green_yellow_color = "b2b200"

        p1 = sd.participant("Participant 1", color=green_color)
        p2 = sd.participant("Participant 2", color=another_green_color)
        p3 = sd.participant("Participant 3")

        sd.group_participants("A first\ngroup", p1, p2, color=green_yellow_color)
        sd.group_participants("A second\ngroup", p3)

        p4 = sd.participant("Participant 4")
        p5 = sd.participant("Participant 5")

        p1.go_to(p2, "Pass a message").go_to(p3, "Pass a message").go_to(
            p4, "Pass a message"
        ).go_to(p5, "Message!")

        sd.group_participants("A third\ngroup", p4, p5)
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Empty Transitions between Participants
participant p1 as First
participant p2 as Second
activate p1
p1->>p2: 
activate p2
p2-->>p1: 
deactivate p2
deactivate p1
""",
            ),
            (
                PlantUML,
                """@startuml
title: Empty Transitions between Participants
participant "First" as p1 
participant "Second" as p2 
activate p1 
p1->p2: 
activate p2 
p2-->p1: 
deactivate p2
deactivate p1
@enduml
""",
            ),
            (
                D2,
                """title: Empty Transitions between Participants {
shape: sequence_diagram
p1: First 
p2: Second 
p1.0 -> p2.1: ''
p2.1 -> p1.0: '' {style.stroke-dash: 3}
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Empty Transitions between Participants
participant "First" as p1
participant "Second" as p2
activate p1
p1->p2: 
activate p2
p2-->p1: 
deactivate p2
deactivate p1
""",
            ),
        ),
    )
    def test_empty_transitions(self, generator_cls, output):
        sd = SequenceDiagram("Empty Transitions between Participants", generator_cls)
        first = sd.participant("First")
        second = sd.participant("Second")

        first.go_to(second).return_to(first)
        assert str(sd) == output

    @pytest.mark.parametrize(
        "generator_cls,output",
        (
            (
                Mermaid,
                """sequenceDiagram
Title: Participant types, according to ECB
actor p1 as Actor
participant p2 as Boundary
participant p3 as Control
participant p4 as Entity
activate p1
p1->>p2: Do something
activate p2
p2->>p3: Do something
activate p3
p3->>p4: Do something
activate p4
p4-->>p3: Return
deactivate p4
p3-->>p2: Return
deactivate p3
p2-->>p1: Return
deactivate p2
deactivate p1
""",
            ),
            (
                PlantUML,
                """@startuml
title: Participant types, according to ECB
actor "Actor" as p1 #769D8F
boundary "Boundary" as p2 
control "Control" as p3 #769D8F
entity "Entity" as p4 
activate p1 
p1->p2: Do something
activate p2 
p2->p3: Do something
activate p3 
p3->p4: Do something
activate p4 
p4-->p3: Return
deactivate p4
p3-->p2: Return
deactivate p3
p2-->p1: Return
deactivate p2
deactivate p1
@enduml
""",
            ),
            (
                D2,
                """title: Participant types, according to ECB {
shape: sequence_diagram
p1: Actor {
style: {fill: "#769D8F" 
stroke:"#769D8F" }
shape: person
}
p2: Boundary 
p3: Control {
style: {fill: "#769D8F" 
stroke:"#769D8F" }
}
p4: Entity 
p1.0 -> p2.1: Do something
p2.1 -> p3.2: Do something
p3.2 -> p4.3: Do something
p4.3 -> p3.2: Return {style.stroke-dash: 3}
p3.2 -> p2.1: Return {style.stroke-dash: 3}
p2.1 -> p1.0: Return {style.stroke-dash: 3}
}
""",
            ),
            (
                SequenceDiagramOrg,
                """title Participant types, according to ECB
actor "Actor" as p1#769D8F
boundary "Boundary" as p2
control "Control" as p3#769D8F
entity "Entity" as p4
activate p1
p1->p2: Do something
activate p2
p2->p3: Do something
activate p3
p3->p4: Do something
activate p4
p4-->p3: Return
deactivate p4
p3-->p2: Return
deactivate p3
p2-->p1: Return
deactivate p2
deactivate p1
""",
            ),
        ),
    )
    def test_ecb_types_for_participants(self, generator_cls, output):
        sd = SequenceDiagram("Participant types, according to ECB", generator_cls)
        actor = sd.participant("Actor", color="769D8F").as_actor()
        boundary = sd.participant("Boundary").as_boundary()
        control = sd.participant("Control", color="769D8F").as_control()
        entity = sd.participant("Entity").as_entity()

        actor.go_to(boundary, "Do something").go_to(control, "Do something").go_to(
            entity, "Do something"
        )
        entity.return_to(control, "Return").return_to(boundary, "Return").return_to(
            actor, "Return"
        )
        assert str(sd) == output

    def test_invalid_color_string(self):
        sd = SequenceDiagram("Invalid color", Mock)
        with pytest.raises(ChartingException):
            sd.participant("Invalid color", "red")

    def test_same_participants(self):
        sd = SequenceDiagram("Same participants", Mock)

        sd.participant("Same")
        with pytest.raises(ChartingException):
            sd.participant(
                "Same"
            )  # it is forbidden to have multiple participants with the same label

    def test_forbid_return_when_not_auto_activated(self):
        sd = SequenceDiagram(
            "Return without auto-activation", Mock, auto_activation=False
        )

        first = sd.participant("First")
        second = sd.participant("Second")

        first.go_to(second)
        with pytest.raises(ChartingException):
            sd.return_()

    def test_forbid_return_to_nowhere(self):
        sd = SequenceDiagram("Return to nowhere", Mock)

        with pytest.raises(ChartingException):
            sd.return_()

    def test_force_case_after_condition(self):
        sd = SequenceDiagram("Violated order of condition and case #1", Mock)
        first = sd.participant("First")
        second = sd.participant("Second")

        with pytest.raises(ChartingException):
            with sd.condition():
                first.go_to(second, "Do smth")

    def test_force_case_inside_condition(self):
        sd = SequenceDiagram("Violated order of condition and case #2", Mock)

        with pytest.raises(ChartingException):
            with sd.case("A case"):
                pass

    @pytest.mark.parametrize(
        "callable_", ("as_actor", "as_boundary", "as_control", "as_entity")
    )
    def test_force_participant_type_being_set_only_once(self, callable_):
        sd = SequenceDiagram("Participant type can be set only once", Mock)

        a = sd.participant("Someone")
        a.type_ = "SET TO SOMETHING ELSE BUT NOT DEFAULT"
        with pytest.raises(ChartingException):
            getattr(a, callable_)()

    @pytest.mark.parametrize(
        "from_,to_",
        (
            ("actor", "actor"),
            ("actor", "control"),
            ("actor", "entity"),
            ("boundary", "boundary"),
            ("boundary", "entity"),
            ("control", "actor"),
            ("entity", "actor"),
            ("entity", "boundary"),
        ),
    )
    def test_only_certain_forward_interactions_are_available(self, from_, to_):
        sd = SequenceDiagram("Not available interactions", Mock)

        a = sd.participant("a")
        b = sd.participant("b")

        a.type_ = from_
        b.type_ = to_

        with pytest.raises(ChartingException):
            a.go_to(b, "Do something")

    @pytest.mark.parametrize(
        "from_,to_",
        (
            ("actor", "actor"),
            ("actor", "control"),
            ("actor", "entity"),
            ("boundary", "boundary"),
            ("boundary", "entity"),
            ("control", "actor"),
            ("entity", "actor"),
            ("entity", "boundary"),
        ),
    )
    def test_only_certain_return_interactions_are_available(self, from_, to_):
        sd = SequenceDiagram("Not available interactions", Mock)

        a = sd.participant("a")
        b = sd.participant("b")

        a.type_ = from_
        b.type_ = to_

        with pytest.raises(ChartingException):
            a.return_to(b, "Do something")

    @pytest.mark.parametrize("title", (None, ""))
    def test_cannot_group_participants_under_group_with_empty_name(self, title):
        sd = SequenceDiagram(
            "Violated participants group naming - empty names not allowed", Mock
        )
        first = sd.participant("First")
        second = sd.participant("Second")
        with pytest.raises(ChartingException):
            sd.group_participants(title, first, second)

    def test_cannot_use_same_group_names_multiple_times(self):
        sd = SequenceDiagram(
            "Violated participants group naming - group names must be unique", Mock
        )
        first = sd.participant("First")
        second = sd.participant("Second")
        group_title = Mock()
        sd.group_participants(group_title, first)
        with pytest.raises(ChartingException):
            sd.group_participants(group_title, second)

    def test_cannot_put_same_participant_to_different_groups(self):
        sd = SequenceDiagram(
            "Violated participants group conditions - participant can be only in one group",
            Mock,
        )
        first = sd.participant("First")
        group_title_1 = Mock()
        group_title_2 = Mock()
        sd.group_participants(group_title_1, first)
        with pytest.raises(ChartingException):
            sd.group_participants(group_title_2, first)
